if (window.IS_DEV) {
  const source = new EventSource('/esbuild');
  source.addEventListener('change', () => location.reload());
}

const xmlns = 'http://www.w3.org/2000/svg';
const svg = document.querySelector('svg');

const hex = {
  inradius: 8.66,
  circumradius: 10,
}

const horzSpacing = hex.inradius;
const vertSpacing = hex.circumradius * 3 / 2;


function toKey({ q, r, s }) {
  return `${[q, r, s]}`;
}

function sameSigns(a, b) {
  return a > -1 && b > -1 || a < 0 && b < 0;
}

function getNeighbors(coords) {
  const [q, r, s] = coords.split(',').map(n => +n);

  return [
    toKey({ q: q + 1, r: r, s: s - 1 }),
    toKey({ q: q - 1, r: r, s: s + 1 }),
    toKey({ q: q + 1, r: r - 1, s: s }),
    toKey({ q: q - 1, r: r + 1, s: s }),
    toKey({ q: q, r: r + 1, s: s - 1 }),
    toKey({ q: q, r: r - 1, s: s + 1 }),
  ]
}

function generateRadialCoordsRect({ rows, columns, odd = false, equal = true } = {}, { q = 0, r = 0, s = 0 } = {}) {
  const list = new Map();

  for (let row = 0; row < rows; row++) {
    const alternating = row % 2;
    const dr = row;

    for (let col = 0; col < columns; col++) {
      let dq = -(alternating + row) / 2 + alternating - col;
      let ds = -(row + alternating) / 2 + col;

      if (odd) {
        dq = dq - alternating;
        ds = ds + alternating;
      }

      list.set(toKey({ q: q + dq, r: r + dr, s: s + ds }), null);
    }

    if (!equal) {
      if ((odd && !(row % 2)) || (!odd && row % 2)) {
        let dq = -(alternating + row) / 2 + alternating - columns;
        let ds = -(row + alternating) / 2 + columns;

        if (odd) {
          dq = dq - alternating;
          ds = ds + alternating;
        }

        list.set(toKey({ q: q + dq, r: r + dr, s: s + ds }), null);
      }
    }
  }

  return list;
}

function addHexText(q, r, s, v) {
  const qText = document.createElementNS(xmlns, 'text');
  qText.textContent = q;
  qText.setAttributeNS(null, 'x', -3);
  qText.setAttributeNS(null, 'y', -3);

  const rText = document.createElementNS(xmlns, 'text');
  rText.textContent = r;
  rText.setAttributeNS(null, 'x', 5);
  rText.setAttributeNS(null, 'y', 1.5);

  const sText = document.createElementNS(xmlns, 'text');
  sText.textContent = s;
  sText.setAttributeNS(null, 'x', -3);
  sText.setAttributeNS(null, 'y', 5);

  const vText = document.createElementNS(xmlns, 'text');
  vText.textContent = v;
  vText.style.fill = 'red';
  vText.setAttributeNS(null, 'y', 1);
  vText.setAttributeNS(null, 'x', -2);

  return [qText, rText, sText, vText];
}

function radialToScreenCoords({ q, r, s }) {
  let x;

  if (q === s)
    x = 0;
  else if (sameSigns(q, s))
    x = Math.abs(q - s);
  else
    x = Math.abs(q) + Math.abs(s);

  x = (q > s ? -1 : 1) * x * horzSpacing;
  y = r * vertSpacing;

  return { x, y };
}

function drawHexes(el, list, renderText = false) {
  for ([key, v] of list) {
    const [q, r, s] = key.split(',').map(n => +n);
    const { x, y } = radialToScreenCoords({ q, r, s });

    const cell = document.createElementNS(xmlns, 'g');
    cell.setAttributeNS(null, 'transform', `translate(${x}, ${y})`);

    const use = document.createElementNS(xmlns, 'use');
    use.setAttributeNS(null, 'href', '#hex');
    cell.appendChild(use);

    if (renderText) addHexText(q, r, s, v).forEach(txt => cell.appendChild(txt));

    el.appendChild(cell);
  }
}

function translateRadialCoords({ q, r, s }, direction, distance = 1) {
  return {
    q: q + direction.q * distance,
    r: r + direction.r * distance,
    s: s + direction.s * distance
  };
}


function generateRadialCoords(l, { q, r, s }, { left, top, right, bottom }, offset) {
  const origin = toKey({ q, r, s });
  const list = new Map(l);

  list.set(origin, 0);
  let queue = [origin];

  while (queue.length > 0) {
    const v = queue.shift();

    getNeighbors(v).forEach(w => {
      const [wq, wr, ws] = w.split(',').map(n => +n);
      const rDist = Math.abs(wr - r);
      const alternating = rDist % 2;
      const dr = (rDist + alternating) / 2;

      let dLeft = left;
      let dRight = right;

      if (['left', 'both'].includes(offset)) dRight -= alternating;
      if (['right', 'both'].includes(offset)) dLeft -= alternating;

      if ([
        !list.has(w),
        wr < bottom + r && wr > -top + r,
        wq > -dRight + q - dr && wq < dLeft + q + dr,
        ws > -dLeft + s - dr && ws < dRight + s + dr,
      ].every(v => v)) {
        list.set(w, dr);
        queue.push(w);
      }
    });
  }

  return list;
}

function translateCoords(map, translator) {
  const translated = new Map();

  for ([key, val] of map) {
    const [q, r, s] = key.split(',').map(n => +n);
    translated.set(toKey(translator({ q, r, s })), val);
  }

  return translated;
}

function placeBuilding(buildingSelector, mapsheetSelector) {
  const building = svg.querySelector(buildingSelector);
  const bld2grid = document.createElementNS(xmlns, 'g');
  const bld2origin = { q: 0, r: 0, s: 0 };

  const bld2map2 = svg.querySelector(`${mapsheetSelector} ${buildingSelector}`);
  const [q, r, s] = bld2map2.dataset.placement.split(',').map(n => +n);
  const map2bld2place = { q, r, s };

  bld2hexes = building.querySelector('[data-grid-footprint]').dataset.gridFootprint.split(':').map(coords => {
    const [q, r, s] = coords.split(',').map(n => +n);
    const { q: tq, r: tr, s: ts } = translateRadialCoords({ q, r, s }, map2bld2place);

    return [toKey({ q: tq, r: tr, s: ts }), null];
  });

  const bld2screenCoords = radialToScreenCoords(translateRadialCoords(bld2origin, map2bld2place));
  const bld2structure = document.createElementNS(xmlns, 'g');
  bld2structure.setAttributeNS(null, 'transform', `translate(${bld2screenCoords.x}, ${bld2screenCoords.y})`)

  building.querySelectorAll('[data-grid-footprint] > *').forEach(el => {
    const use = document.createElementNS(xmlns, 'use');
    use.setAttributeNS(null, 'href', `#${el.id}`);
    el.classList.forEach(className => use.classList.add(className));
    bld2structure.appendChild(use);
  });

  drawHexes(bld2grid, new Map(bld2hexes), true);

  const bld2elvBasement = bld2grid.cloneNode(true);
  bld2elvBasement.classList.add('elevation-basement');
  bld2map2.appendChild(bld2elvBasement);

  const bld2elv1 = bld2grid.cloneNode(true);
  bld2elv1.classList.add('elevation-1');
  bld2map2.appendChild(bld2elv1);

  const bld2elv2 = bld2grid.cloneNode(true);
  bld2elv2.classList.add('elevation-2');
  bld2map2.appendChild(bld2elv2);

  const bld2elvRoof = bld2grid.cloneNode(true);
  bld2elvRoof.classList.add('elevation-roof');
  bld2map2.appendChild(bld2elvRoof);

  bld2map2.appendChild(bld2structure);
}

// const building1 = generateRadialCoordsRect({ rows: 5, columns: 3, equal: false }, { q: 7, r: -4, s: -3 });
// const building2 = generateRadialCoordsRect({ rows: 4, columns: 3, odd: true }, { q: 0, r: -4, s: 4 });
// const building3 = generateRadialCoords(new Map(), { q: -1, r: 2, s: -1 }, 1);


// const buildingCoords = [
//   generateRadialCoordsRect({ rows: 4, columns: 8, equal: false }, { q: 18, r: -8, s: -10 }),
//   generateRadialCoordsRect({ rows: 8, columns: 2, equal: false }, { q: 9, r: -10, s: 1 }),
//   generateRadialCoordsRect({ rows: 6, columns: 13, odd: true }, { q: 2, r: -8, s: 6 }),
//   generateRadialCoordsRect({ rows: 1, columns: 10 }, { q: -1, r: -9, s: 10 }),
//   generateRadialCoordsRect({ rows: 1, columns: 2 }, { q: -1, r: -2, s: 3 }),
//   generateRadialCoordsRect({ rows: 1, columns: 6 }, { q: -8, r: -2, s: 10 }),
//   // generateRadialCoordsRect({ rows: 8, columns: 4, odd: true }, { q: 4, r: 3, s: -7 }),
//   generateRadialCoordsRect({ rows: 8, columns: 4, odd: true }, { q: -5, r: 3, s: 2 }),
//   generateRadialCoordsRect({ rows: 6, columns: 3, equal: false }, { q: -12, r: 3, s: 9 }),

//   // hex building
//   // generateRadialCoordsRect({ rows: 8, columns: 4 }, { q: 10, r: 3, s: -13 }),
//   // generateRadialCoordsRect({ rows: 3, columns: 1 }, { q: 9, r: 7, s: -16 }),
//   // generateRadialCoordsRect({ rows: 5, columns: 1, odd: true }, { q: 5, r: 6, s: -11 }),
// ]

// building2.forEach(coords => grid.delete(coords));
// building3.forEach(coords => grid.delete(coords));

// drawHexes(svg, generateRadialCoords(new Map(), { q: 0, r: 0, s: 0 }, 16));

// const rectEven = generateRadialCoordsRect({ rows: 2, columns: 2 }, { q: 9, r: -5, s: -4 });
// const rectOdd = generateRadialCoordsRect({ rows: 2, columns: 2, odd: true }, { q: 8, r: -2, s: -6 });
// const rectEvenUnevenCols = generateRadialCoordsRect({ rows: 2, columns: 1, equal: false}, { q: 6, r: 1, s: -7 });
// const rectOddUnevenCols = generateRadialCoordsRect({ rows: 2, columns: 1, odd: true, equal: false}, { q: 5, r: 4, s: -9 });

// drawHexes(svg, rectEven);
// drawHexes(svg, rectOdd);
// drawHexes(svg, rectEvenUnevenCols);
// drawHexes(svg, rectOddUnevenCols);

// drawHexes(svg, generateRadialCoordsRect({ rows: 7, columns: 6, odd: true, equal: false }, { q: 5, r: -5, s: 0 }));

// const buildings = document.createElementNS(xmlns, 'g');
// buildings.classList.add('buildings');

// const bldElevationBasement = document.createElementNS(xmlns, 'g');
// bldElevationBasement.classList.add('elevation-basement');
// buildingCoords.forEach(building => drawHexes(bldElevationBasement, building));

// const bldElevation1 = document.createElementNS(xmlns, 'g');
// bldElevation1.classList.add('elevation-1');
// buildingCoords.forEach(building => drawHexes(bldElevation1, building));

// const bldElevation2 = document.createElementNS(xmlns, 'g');
// bldElevation2.classList.add('elevation-2');
// buildingCoords.forEach(building => drawHexes(bldElevation2, building));

// buildings.appendChild(bldElevationBasement);
// buildings.appendChild(bldElevation1);
// buildings.appendChild(bldElevation2);

const elevation1 = document.createElementNS(xmlns, 'g');
elevation1.classList.add('elevation-1');

const elevation2 = document.createElementNS(xmlns, 'g');
elevation2.classList.add('elevation-2');

const buildingHexes = {};

buildingHexes.bld2 = generateRadialCoords(
  new Map(),
  { q: 0, r: 0, s: 0 },
  { left: 3, top: 5, right: 3, bottom: 5 },
  'both'
);

buildingHexes.bld2 = generateRadialCoords(
  buildingHexes.bld2,
  translateRadialCoords({ q: 0, r: 0, s: 0 }, { q: 1, r: 0, s: -1 }, 3),
  { left: 1, top: 2, right: 1, bottom: 4 },
  'right'
);

buildingHexes.bld2 = generateRadialCoords(
  buildingHexes.bld2,
  translateRadialCoords({ q: 0, r: 0, s: 0 }, { q: 1, r: 0, s: -1 }, -3),
  { left: 1, top: 2, right: 1, bottom: 4 },
  'left'
);

buildingHexes.bld1 = generateRadialCoords(
  new Map(),
  { q: 0, r: 0, s: 0 },
  { left: 3, top: 5, right: 3, bottom: 5 },
  'both'
);

buildingHexes.bld3 = generateRadialCoords(
  new Map(),
  { q: 0, r: 0, s: 0 },
  { left: 2, top: 3, right: 2, bottom: 4 }
);

buildingHexes.bld4 = generateRadialCoords(
  new Map(),
  { q: 0, r: 0, s: 0 },
  { left: 2, top: 4, right: 3, bottom: 5 },
  'left'
);

buildingHexes.bld5 = generateRadialCoords(
  new Map(),
  { q: 0, r: 0, s: 0 },
  { left: 5, top: 3, right: 4, bottom: 2 }
);

buildingHexes.bld6 = generateRadialCoords(
  new Map(),
  { q: 0, r: 0, s: 0 },
  { left: 1, top: 5, right: 2, bottom: 4 }
);

buildingHexes.bld7 = generateRadialCoords(
  new Map(),
  { q: 0, r: 0, s: 0 },
  { left: 7, top: 4, right: 7, bottom: 3 },
  'left'
);

buildingHexes.bld7 = generateRadialCoords(
  buildingHexes.bld7,
  { q: 1, r: -4, s: 3 },
  { left: 5, top: 1, right: 6, bottom: 1 }
);

buildingHexes.bld7 = generateRadialCoords(
  buildingHexes.bld7,
  { q: 5, r: 3, s: -8 },
  { left: 1, top: 1, right: 2, bottom: 1 }
);

buildingHexes.bld7 = generateRadialCoords(
  buildingHexes.bld7,
  { q: -4, r: 3, s: 1 },
  { left: 3, top: 1, right: 4, bottom: 1 }
);

const grid = generateRadialCoordsRect({ rows: 25, columns: 33 }, { q: 22, r: -12, s: -10 });
// const grid = generateRadialCoordsRect({ rows: 52, columns: 33 }, translateRadialCoords({ q: 22, r: -12, s: -10 }, { q: 1, r: -2, s: 1 }, 7));

const mapsheet1BuildingData = [
  [{ rows: 6, columns: 3, equal: false }, { q: -4, r: -8, s: 12 }],
  [{ rows: 8, columns: 2, equal: false, odd: true }, { q: -10, r: 4, s: 6 }],

  [{ rows: 9, columns: 4, odd: true, equal: false }, { q: 11, r: 3, s: -14 }],
  [{ rows: 5, columns: 1 }, { q: 10, r: 6, s: -16 }],
  [{ rows: 5, columns: 1, odd: true }, { q: 5, r: 6, s: -11 }],

  [{ rows: 8, columns: 4 }, { q: 17, r: -10, s: -7 }],

  // tree/brush copse
  [{ rows: 9, columns: 6, equal: false }, { q: 5, r: -4, s: -1 }],
];

const mapsheet1BuildingCoords = mapsheet1BuildingData.map(([size, start]) => generateRadialCoordsRect(size, start));

const mapsheet2BuildingCoords = [
  translateCoords(buildingHexes.bld1, ({ q, r, s }) => ({ q: q, r: r + 7, s: s - 7 })),
  translateCoords(buildingHexes.bld2, ({ q, r, s }) => ({ q: q + 7, r: r + 7, s: s - 14 })),
  translateCoords(buildingHexes.bld3, ({ q, r, s }) => ({ q: q - 14, r: r + 5, s: s + 9 })),
  translateCoords(buildingHexes.bld4, ({ q, r, s }) => ({ q: q - 8, r: r + 6, s: s + 2 })),
  translateCoords(buildingHexes.bld5, ({ q, r, s }) => ({ q: q + 13, r: r - 6, s: s - 7 })),
  translateCoords(buildingHexes.bld6, ({ q, r, s }) => ({ q: q + 7, r: r - 6, s: s - 1 })),
  translateCoords(buildingHexes.bld7, ({ q, r, s }) => ({ q: q - 6, r: r - 5, s: s + 11 })),
];

const mapsheet3BuildingCoords = [
  translateCoords(buildingHexes.bld1, ({ q, r, s }) => ({ q: q - 2, r: r - 5, s: s + 7 })),
  translateCoords(buildingHexes.bld2, ({ q, r, s }) => ({ q: -s + 9, r: -r - 6, s: -q - 3 })),
  translateCoords(buildingHexes.bld3, ({ q, r, s }) => ({ q: -s + 17, r: -r - 7, s: -q - 10 })),
  translateCoords(buildingHexes.bld4, ({ q, r, s }) => ({ q: q - 9, r: r - 6, s: s + 15 })),
  translateCoords(buildingHexes.bld5, ({ q, r, s }) => ({ q: q + 6, r: r + 8, s: s - 14 })),
  translateCoords(buildingHexes.bld6, ({ q, r, s }) => ({ q: -s + 1, r: -r + 6, s: -q - 7 })),
  translateCoords(buildingHexes.bld7, ({ q, r, s }) => ({ q: q - 12, r: r + 8, s: s + 4 })),
]

// drawHexes(elevation2, translateCoords(buildingHexes.bld7, ({ q, r, s }) => ({ q: q - 12, r: r + 8, s: s + 4 })), true);
// drawHexes(elevation1, buildingHexes.bld7, true);

const mapsheetHexCoords = generateRadialCoords(new Map(), { q: 0, r: 0, s: 0 }, { left: 17, top: 13, right: 17, bottom: 14 });

const mapsheet1hexCoords = new Map(mapsheetHexCoords);
const mapsheet2hexCoords = new Map(mapsheetHexCoords);
const mapsheet3hexCoords = new Map(mapsheetHexCoords);

mapsheet2BuildingCoords.forEach(building => {
  for ([coords, v] of building) mapsheet2hexCoords.delete(coords);
});

mapsheet3BuildingCoords.forEach(building => {
  for ([coords, v] of building) mapsheet3hexCoords.delete(coords);
});

function coordsScenario2() {
  const grid = generateRadialCoords(new Map(), { q: 0, r: 0, s: 0 }, { left: 17, top: 39, right: 17, bottom: 40 });

  // mapsheet2BuildingCoords
  //   // .map(([size, start]) => [size, translateRadialCoords(start, { q: 1, r: -2, s: 1 }, 13)])
  //   .map(([size, start]) => [size, translateRadialCoords(start, { q: 1, r: -2, s: 1 }, 13)])
  //   .map(([size, start]) => generateRadialCoordsRect(size, start)).forEach(building => {
  //     for ([coords, v] of building) grid.delete(coords);
  //   }
  // );

  // mapsheet3BuildingCoords
  //   .map(([size, start]) => [size, translateRadialCoords(start, { q: 1, r: -2, s: 1 }, -13)])
  //   .map(([size, start]) => generateRadialCoordsRect(size, start)).forEach(building => {
  //     for ([coords, v] of building) grid.delete(coords);
  //   }
  // );

  // mapsheet1BuildingData
  //   .map(([size, start]) => generateRadialCoordsRect(size, start)).forEach(building => {
  //     for ([coords, v] of building) grid.delete(coords);
  //   }
  // );

  return new Map([...mapsheet2hexCoords]);
}

// drawHexes(elevation1, coordsScenario1(), true);

const gameboard = svg.querySelector('.gameboard');

function drawScenario1() {
  const mapsheet2container = document.createElementNS(xmlns, 'g');
  mapsheet2container.id = 'mapsheet2';
  gameboard.appendChild(mapsheet2container);

  const mapsheet3container = document.createElementNS(xmlns, 'g');
  mapsheet3container.id = 'mapsheet3';
  gameboard.appendChild(mapsheet3container);

  const [map2elevation1, map3elevation1] = [mapsheet2container, mapsheet3container].map(el => {
    const hexContainer = document.createElementNS(xmlns, 'g');
    hexContainer.classList.add('elevation-1');
    el.appendChild(hexContainer);
    return hexContainer;
  });

  [
    [map2elevation1, mapsheet2hexCoords, ({ q, r, s }) => ({ q: q + 7, r: r - 14, s: s + 7 })],
    [map3elevation1, mapsheet3hexCoords, ({ q, r, s }) => ({ q: q - 6, r: r + 12, s: s - 6 })],
  ].forEach(([el, coords, translator]) => drawHexes(el, translateCoords(coords, translator), true));

  mapsheet2BuildingCoords.forEach((buildingCoords, index) => {
    const buildingContainer = document.createElementNS(xmlns, 'g');
    buildingContainer.classList.add(`building`);
    buildingContainer.classList.add(`building${index}`);
    mapsheet2container.appendChild(buildingContainer);

    const translatedCoords = translateCoords(buildingCoords, ({ q, r, s }) =>
      ({ q: q + 7, r: r - 14, s: s + 7 })
    );

    ['elevation-1', 'elevation-2', 'elevation-basement'].forEach(elevation => {
      const hexContainer = document.createElementNS(xmlns, 'g');
      hexContainer.classList.add(elevation);
      buildingContainer.appendChild(hexContainer);

      drawHexes(hexContainer, translatedCoords, true);
    });
  });

  mapsheet3BuildingCoords.forEach((buildingCoords, index) => {
    const buildingContainer = document.createElementNS(xmlns, 'g');
    buildingContainer.classList.add(`building`);
    buildingContainer.classList.add(`building${index}`);
    mapsheet3container.appendChild(buildingContainer);

    const translatedCoords = translateCoords(buildingCoords, ({ q, r, s }) =>
      ({ q: q - 6, r: r + 12, s: s - 6 })
    );

    ['elevation-1', 'elevation-2', 'elevation-basement'].forEach(elevation => {
      const hexContainer = document.createElementNS(xmlns, 'g');
      hexContainer.classList.add(elevation);
      buildingContainer.appendChild(hexContainer);

      drawHexes(hexContainer, translatedCoords, true);
    });
  });
}

// const gameboardGrid = svg.querySelector('.grid');

drawScenario1();

// drawHexes(document.querySelector('#mapsheet1'), mapsheet1, true);

// placeBuilding('.building2', '#mapsheet2');
// drawHexes(document.querySelector('#mapsheet2 .grid'), mapsheet2, true);

// placeBuilding('.building2', '#mapsheet3');
// drawHexes(document.querySelector('#mapsheet3 .grid'), mapsheet3, true);


// defs.after(elevation2);
// gameboardGrid.appendChild(elevation1);
// gameboardGrid.appendChild(elevation2);
